/* 
 * files.c -- DHCP server file manipulation *
 * Rewrite by Russ Dill <Russ.Dill@asu.edu> July 2001
 */
 
#include <stdio.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <ctype.h>
#include <netdb.h>

#include "debug.h"
#include "dhcpd.h"
#include "files.h"
#include "options.h"
#include "leases.h"
//storm
#include "gigaconfig.h"
#include "libgigarouter.h"

/* on these functions, make sure you datatype matches */
static int read_ip(char *line, void *arg)
{
	struct in_addr *addr = arg;
	struct hostent *host;
	int retval = 1;

	if (!inet_aton(line, addr)) {
		if ((host = gethostbyname(line))) 
			addr->s_addr = *((unsigned long *) host->h_addr_list[0]);
		else retval = 0;
	}
	return retval;
}


static int read_str(char *line, void *arg)
{
	char **dest = arg;
	
	if (*dest) free(*dest);
	*dest = strdup(line);
	
	return 1;
}


static int read_u32(char *line, void *arg)
{
	u_int32_t *dest = arg;
	char *endptr;
	*dest = strtoul(line, &endptr, 0);
	return endptr[0] == '\0';
}


static int read_yn(char *line, void *arg)
{
	char *dest = arg;
	int retval = 1;

	if (!strcasecmp("yes", line))
		*dest = 1;
	else if (!strcasecmp("no", line))
		*dest = 0;
	else retval = 0;
	
	return retval;
}


/* read a dhcp option and add it to opt_list */
static int read_opt(char *line, void *arg)
{
	struct option_set **opt_list = arg;
	char *opt, *val, *endptr;
	struct dhcp_option *option = NULL;
	int retval = 0, length = 0;
	char buffer[255];
	u_int16_t result_u16;
	u_int32_t result_u32;
	int i;

	if (!(opt = strtok(line, " \t="))) return 0;
	
	for (i = 0; options[i].code; i++)
		if (!strcmp(options[i].name, opt))
			option = &(options[i]);
		
	if (!option) return 0;
	
	do {
		val = strtok(NULL, ", \t");
		if (val) {
			length = option_lengths[option->flags & TYPE_MASK];
			retval = 0;
			switch (option->flags & TYPE_MASK) {
			case OPTION_IP:
				retval = read_ip(val, buffer);
				break;
			case OPTION_IP_PAIR:
				retval = read_ip(val, buffer);
				if (!(val = strtok(NULL, ", \t/-"))) retval = 0;
				if (retval) retval = read_ip(val, buffer + 4);
				break;
			case OPTION_STRING:
				length = strlen(val);
				if (length > 0) {
					if (length > 254) length = 254;
					memcpy(buffer, val, length);
					retval = 1;
				}
				break;
			case OPTION_BOOLEAN:
				retval = read_yn(val, buffer);
				break;
			case OPTION_U8:
				buffer[0] = strtoul(val, &endptr, 0);
				retval = (endptr[0] == '\0');
				break;
			case OPTION_U16:
				result_u16 = htons(strtoul(val, &endptr, 0));
				memcpy(buffer, &result_u16, 2);
				retval = (endptr[0] == '\0');
				break;
			case OPTION_S16:
				result_u16 = htons(strtol(val, &endptr, 0));
				memcpy(buffer, &result_u16, 2);
				retval = (endptr[0] == '\0');
				break;
			case OPTION_U32:
				result_u32 = htonl(strtoul(val, &endptr, 0));
				memcpy(buffer, &result_u32, 4);
				retval = (endptr[0] == '\0');
				break;
			case OPTION_S32:
				result_u32 = htonl(strtol(val, &endptr, 0));	
				memcpy(buffer, &result_u32, 4);
				retval = (endptr[0] == '\0');
				break;
			default:
				break;
			}
			if (retval) 
				attach_option(opt_list, option, buffer, length);
		};
	} while (val && retval && option->flags & OPTION_LIST);
	return retval;
}


static struct config_keyword keywords[] = {
	/* keyword[14]	handler   variable address		default[20] */
	{"start",	read_ip,  &(server_config.start),	"192.168.0.20"},
	{"end",		read_ip,  &(server_config.end),		"192.168.0.254"},
	{"interface",	read_str, &(server_config.interface),	"eth0"},
	{"option",	read_opt, &(server_config.options),	""},
	{"opt",		read_opt, &(server_config.options),	""},
	{"max_leases",	read_u32, &(server_config.max_leases),	"254"},
	{"remaining",	read_yn,  &(server_config.remaining),	"yes"},
	{"auto_time",	read_u32, &(server_config.auto_time),	"7200"},
	{"decline_time",read_u32, &(server_config.decline_time),"3600"},
	{"conflict_time",read_u32,&(server_config.conflict_time),"3600"},
	{"offer_time",	read_u32, &(server_config.offer_time),	"60"},
	{"min_lease",	read_u32, &(server_config.min_lease),	"60"},
	{"lease_file",	read_str, &(server_config.lease_file),	"/ramdisk/etc/udhcpd.leases"},
	{"pidfile",	read_str, &(server_config.pidfile),	"/var/run/udhcpd.pid"},
	{"notify_file", read_str, &(server_config.notify_file),	""},
	{"siaddr",	read_ip,  &(server_config.siaddr),	"0.0.0.0"},
	{"sname",	read_str, &(server_config.sname),	""},
	{"boot_file",	read_str, &(server_config.boot_file),	""},
	{"dnsrelay",	read_u32, &(server_config.dnsrelay_enable),"1"},
	/*ADDME: static lease */
	{"",		NULL, 	  NULL,				""}
};


int read_config(char *file)
{
	FILE *in;
	char buffer[80], orig[80], *token, *line;
	int i;

	for (i = 0; strlen(keywords[i].keyword); i++)
		if (strlen(keywords[i].def))
			keywords[i].handler(keywords[i].def, keywords[i].var);

	if (!(in = fopen(file, "r"))) {
		LOG(LOG_ERR, "unable to open config file: %s", file);
		return 0;
	}
	
	while (fgets(buffer, 80, in)) {
		if (strchr(buffer, '\n')) *(strchr(buffer, '\n')) = '\0';
		strncpy(orig, buffer, 80);
		if (strchr(buffer, '#')) *(strchr(buffer, '#')) = '\0';
		token = buffer + strspn(buffer, " \t");
		if (*token == '\0') continue;
		line = token + strcspn(token, " \t=");
		if (*line == '\0') continue;
		*line = '\0';
		line++;
		
		/* eat leading whitespace */
		line = line + strspn(line, " \t=");
		/* eat trailing whitespace */
//		for (i = strlen(line) - 1; i > 0 && isspace(line[i]); i--);
//		line[i > 0 ? i + 1 : 0] = '\0';
                for (i = strlen(line); i > 0 && isspace(line[i - 1]); i--);
		line[i] = '\0';

		for (i = 0; strlen(keywords[i].keyword); i++)
			if (!strcasecmp(token, keywords[i].keyword))
				if (!keywords[i].handler(line, keywords[i].var)) {
					LOG(LOG_ERR, "unable to parse '%s'", orig);
					/* reset back to the default value */
					keywords[i].handler(keywords[i].def, keywords[i].var);
				}
	}
	fclose(in);
	return 1;
}


/* the dummy var is here so this can be a signal handler */
void write_leases(int dummy)
{
	FILE *fp;
	unsigned int i;
	char buf[255];
	time_t curr = time(0);
	unsigned long lease_time;
	
	dummy = 0;
	
	if (!(fp = fopen(server_config.lease_file, "w"))) {
		LOG(LOG_ERR, "Unable to open %s for writing", server_config.lease_file);
		return;
	}
	
	for (i = 0; i < server_config.max_leases; i++) {
		if (leases[i].yiaddr != 0) {
			if ((leases[i].static_f == 1) &&
			    (leases[i].expires == 0xffffffff) )
			{
				lease_time = leases[i].expires;
			}
			else
			{
				if (server_config.remaining) {
					if (lease_expired(&(leases[i])))
						lease_time = 0;
					else lease_time = leases[i].expires - curr;
				} else lease_time = leases[i].expires;
			}

			lease_time = htonl(lease_time);
			fwrite(leases[i].chaddr, 16, 1, fp);
			fwrite(&(leases[i].yiaddr), 4, 1, fp);
			fwrite(&lease_time, 4, 1, fp);
			fwrite(&(leases[i].static_f), 4, 1, fp);
		}
	}
	fclose(fp);
	
	if (server_config.notify_file) {
		sprintf(buf, "%s %s", server_config.notify_file, server_config.lease_file);
		system(buf);
	}
}


void read_leases(char *file)
{
	FILE *fp;
	unsigned int i = 0;
//	time_t curr = time(0);
	struct dhcpOfferedAddr lease;
	
	if (!(fp = fopen(file, "r"))) {
		LOG(LOG_ERR, "Unable to open %s for reading", file);
		return;
	}
	
	while (i < server_config.max_leases && (fread(&lease, sizeof lease, 1, fp) == 1)) {
		/* ADDME: is it a static lease */
		if (ntohl(lease.yiaddr) >= ntohl(server_config.start) && ntohl(lease.yiaddr) <= ntohl(server_config.end)) {
			lease.expires = ntohl(lease.expires);
//			if (server_config.remaining) lease.expires += curr;
			if (!server_config.remaining) lease.expires -= time(0);
			if (!(add_lease(lease.chaddr, lease.yiaddr, lease.expires))) {
				LOG(LOG_WARNING, "Too many leases while loading %s\n", file);
				break;
			}				
			i++;
		}
	}
	DEBUG(LOG_INFO, "Read %d leases", i);
	fclose(fp);
}
		
void read_static_leases(char *file)
{
	FILE *fp;
	unsigned int i = 0;
	struct dhcpOfferedAddr lease;
	
	if (!(fp = fopen(file, "r"))) {
		LOG(LOG_ERR, "Unable to open %s for reading", file);
		return;
	}

	while (i < server_config.max_leases && (fread(&lease, sizeof lease, 1, fp) == 1)) {
		/* ADDME: is it a static lease */
		if (ntohl(lease.yiaddr) >= ntohl(server_config.start) && ntohl(lease.yiaddr) <= ntohl(server_config.end)) {
			lease.expires = 0xffffffff;
			if (!(add_static_lease(lease.chaddr, lease.yiaddr, lease.expires))) {
				LOG(LOG_WARNING, "Too many leases while loading %s\n", file);
				break;
			}
			i++;
		}
	}
	DEBUG(LOG_INFO, "Read %d leases", i);
	fclose(fp);
}
		
void write_Mleases(char* file)
{
        #define BUFFER_SIZE 64
	
	FILE *fp;
	unsigned int i,j;
	unsigned long lease_time;
	unsigned long INTaddr,INTmac;
	unsigned char chaddr[32]="";
	u_int8_t chaddrt[16]="";
	char szBuffer[BUFFER_SIZE]="";
	char yiaddr[16]="";
	char flag[1]="0";
	char TA[6]="0x";	
	char TA1[6]="0x";
	char TA2[6]="0x";
	char tempstring[128]="";	
	u_int32_t Etime;
	int nResult;
	int nDebug = 0;
	u_int32_t Static_f=0;
	
 //storm / 2002.10.29 // To get static leases into temp lease file 
    		
    nResult = GIGA_GetSystemSetting("DHCP_SERVER", "dhcpd_enable", szBuffer, BUFFER_SIZE, RW_ITEM_MODE);
    if ( nResult != 0 )
    {
        printf("dhcpd: GIGA_GetSystemSetting(DHCP_SERVER, dhcpd_enable, ...) fail, nResult=%d\n", nResult);
        return;
    }
    
    if ( szBuffer[0] == '0' )
    {
        printf("dhcpd: dhcpd is config to disable: exit\n");
        return;
    }
    //---
	
	if (!(fp = fopen( file , "w"))) {
		LOG(LOG_ERR, "Unable to open %s for writing", server_config.lease_file);
		return;
	}
	
	
for (i = 1; i <= 32; i++) {
	
    strcpy(chaddrt,"");
    strcpy(TA,"0x");	
    strcpy(TA1,"0x");
    strcpy(TA2,"0x");
    
    //MAC Lock flags	
    sprintf(tempstring,"dhcpd_flg%d",i);	
    nResult = GIGA_GetSystemSetting("DHCP_SERVER",tempstring, flag, 2, RW_ITEM_MODE);
    if ( nResult != 0 )
    {
        printf("dhcpd: GIGA_GetSystemSetting(DHCP_SERVER, dhcpd_flg%d, ...) fail, nResult=%d\n",i, nResult);
        return;
    }
    if(nDebug)
        printf("dhcpd: dhcpd_flg%d: %s\n",i,flag);

    if(!strcmp(flag,"1")){
    	    
    // IP address
    sprintf(tempstring,"dhcpd_ipa%d",i);    
        nResult = GIGA_GetSystemSetting("DHCP_SERVER", tempstring, yiaddr, 16, RW_ITEM_MODE);
    if ( nResult != 0 ){
        printf("dhcpd: GIGA_GetSystemSetting(DHCP_SERVER, dhcp_ipa%d, ...) fail, nResult=%d\n", i,nResult);
        return;
    }    
    if(nDebug)
        printf("dhcpd: dhcpd_ip%d: %s\n",i, yiaddr); 
        
    // MAC address    
    sprintf(tempstring,"dhcpd_mac%d",i);  
    
        nResult = GIGA_GetSystemSetting("DHCP_SERVER", tempstring, chaddr, 32, RW_ITEM_MODE);
	
    if ( nResult != 0 ){
        printf("dhcpd: GIGA_GetSystemSetting(DHCP_SERVER, dhcp_mac%d, ...) fail, nResult=%d\n",i, nResult);
        return;
    }    
    if(nDebug)
        printf("dhcpd: dhcp_mac%d: %s\n",i, chaddr);         			
   
              if (yiaddr != "") {
      		      
                 lease_time = LEASE_TIME;
		 Etime=lease_time;
	 	 
    //Convert to leases file format.
    //16  byte MAC
    // 4  byte ipaddress
    //u32 expire time
    // 4  byte static flag
    //16  byte MAC
    // 4  byte ipaddress
    //u32 expire time
    // 4  byte static flag ......
                //parsing ":"
		for (j = 0; j < 18; j=j+3){
                     strncat(chaddrt,chaddr+j,2);
		}	
		// write mac to lease file
		        printf("MAC=%s\n",chaddrt);	
		
			strncat(TA,chaddrt,4);
//		        printf("TA=%s\n",TA);	
	             	INTmac=strtol(TA,NULL,16);
                        // rtl8181 use big endian   pjsun
                        INTmac = INTmac << 16;
			
			strncat(TA1,chaddrt+4,4);
//		        printf("TA1=%s\n",TA1);	
	             	INTmac += (strtol(TA1,NULL,16));
			INTmac=htons(INTmac);
		        fwrite(&INTmac, 4, 1, fp);	
				
			strncat(TA2,chaddrt+8,4);
	             	INTmac=strtol(TA2,NULL,16);
                        // rtl8181 use big endian   pjsun
                        INTmac = INTmac << 16;
			INTmac=htons(INTmac);
//		        printf("TA2=%s %04x\n",TA2,INTmac);	
		        fwrite(&INTmac, 4, 1, fp);
		//write space follow MAC	
			INTmac=0x00;	
		        fwrite(&INTmac, 4, 1, fp);	
		        fwrite(&INTmac, 4, 1, fp);	
			
		//write IP to lease file	
			INTaddr=inet_addr(yiaddr);
//			printf("the convert IP is :%lx\n",INTaddr);
			fwrite(&INTaddr, 4, 1, fp);
		//write Time to lease file
//			printf("Time=%d\n",Etime);
	  		fwrite(&Etime, 4, 1, fp);
		//write static flag to lease file
			fwrite(&Static_f, 4, 1, fp);
			
               }//end of IP=NULL.
      	      
	   }// end of flag =="1".
    
	}// end of for() loop.

	fclose(fp);
}		
